import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import javax.sound.midi.*;
import java.util.*;

/**
 * Graphical representation of a MIDI track.
 * 
 * @author  Andersson , Andreen, Lanneskog, Pehrson
 * @version 1
 */

public class MooTrackView extends JPanel {

	private Track track;
	private MooTrackTitle title;
	private Rectangle box;

	private JPopupMenu popup;
	private JMenuItem popupAdd;
	private ArrayList rects;
	private Insets insets;
	private int ticksPerSixteenth, popupY = 0;
	protected static int viewLength = 0;
	protected static int extraHeight = 0;
	public static final int NOTE_HEIGHT = 10, NOTE_WIDTH = 40, VIEW_WIDTH = 200;

	/**
	 * Creates the trackview.
	 * @param track The track it represents graphically and operates on.
	 * @param title	The object that is used to manipulate instrument, channel, solo, mute.
	 */
	public MooTrackView (Track track, MooTrackTitle title) {
		super(true);
		this.track = track;
		this.title = title;
		insets = getInsets();

		// Configures panel
		setBackground(Color.white);
		setBorder(BorderFactory.createLineBorder(Color.black));
		setLayout(null);
		setPreferredSize(new Dimension(VIEW_WIDTH, 140 * NOTE_HEIGHT));

		// Creates temporary variables
		MidiEvent note;
		MooNoteElement elem;
		rects = new ArrayList(track.size() / 2);
		extraHeight = Toolkit.getDefaultToolkit().getScreenSize().height - 150;

		// Places note elements
		for (int i = 0; i < track.size(); i++) {
			note = track.get(i);
			if (note instanceof MooNote) {
				// Adds the note element to the note area and moves it to the appropriate place.
				MooNote mn = (MooNote)note;
				elem = new MooNoteElement(this, mn);
				add(elem);
				layoutElement(elem, false);
			}
			setPreferredSize(new Dimension(VIEW_WIDTH, viewLength + extraHeight));

		}

		// Creates pop-up menu.
		popup = new JPopupMenu();
		PopupListener pList = new PopupListener();
		popupAdd = new JMenuItem("Add note...");
		popupAdd.addActionListener(pList);
		popup.add(popupAdd);

		// Adds listeners for popup menu and keyboard synthesizer.
		addMouseListener(new MAdapter());
		addKeyListener(new MooKeyboard());
	}

	/**
	 * Layouts the element to the right place.
	 * @param elem	the element that will be layouted.
	 * @param old	If true, this method will remove the old layout and set the new preferredSize for the trackview.
	 */
	public void layoutElement(MooNoteElement elem, boolean old){
		// If the element is currently in the view, removes its coordinates from the list.
		Rectangle r = new Rectangle();
		if (old){
			r = elem.getBounds(r);
			for (Iterator i = rects.iterator(); i.hasNext();){
				Object ob = i.next();
				if (r.equals(ob)){
					rects.remove(ob);
					break;
				}
			}
		}

		// Creates temporary variables.
		ticksPerSixteenth = Moosique.getSequence().getResolution() / 4;
		MooNote mn = elem.getNote();
		int x, y, height;

		// Calculates coordinates.
		x = insets.left;
		y = insets.top + (int)(mn.getTick() / ticksPerSixteenth) * NOTE_HEIGHT;
		height = (mn.getDuration() / ticksPerSixteenth) * NOTE_HEIGHT;
		if (height == 0) height = NOTE_HEIGHT;
		r = new Rectangle(x, y, NOTE_WIDTH, height);

		// Places the element in the appropriate place.
		while(isOccupied(r)) r.translate(NOTE_WIDTH, 0);
		elem.setBounds(r);
		rects.add(r);
		if (viewLength < (y + height)){
			viewLength = y + height;
			if(old)setPreferredSize(new Dimension(VIEW_WIDTH, viewLength + extraHeight));
		}
	}

	/** 
	 * Returns the track of this view.
	 * @return the track of this view
	 */
	public Track getTrack() {
		return track;
	}

	/** 
	 * Returns the title of this view.
	 * @return the title of this view
	 */
	public MooTrackTitle getTitle() {
		return title;
	}

	/**
	 * Checks if the element can be fully drawn as this position without inteferring with other elements.
	 * @return true if the position is occupied.
	 */
	private boolean isOccupied(Rectangle r) {
		Iterator it = rects.iterator();
		while (it.hasNext()) {
			if(r.intersects((Rectangle)it.next())) return true;
		}
		return false;
	}
	
	/** 
	 * Adds the given note to the current track, and visualises it.
	 * @param mn	the note to add
	 */
	public void addNote(MooNote mn) {
		mn.addTo(track);
		MooNoteElement elem = new MooNoteElement(this, mn);
		add(elem);
		layoutElement(elem, false);
		setPreferredSize(new Dimension(VIEW_WIDTH, viewLength + extraHeight));
		repaint();
	}

	/** 
	 * Removes the given note element from the view and its note from the current track.
	 * @param elem	the note element to remove
	 */
	public void removeNote(MooNoteElement elem) {
		elem.getNote().removeFrom(track);
		remove(elem);
		elem.getNote().removeFrom(track);
		repaint();
	}

	/**
	 * Adds a standard note to this track.
	 */
	private void addStandardNote() {
		int row =  (popupY - insets.top) / NOTE_HEIGHT;
		long timestamp = (long)(ticksPerSixteenth * row);
		addNote(new MooNote(title.getChannel(), 60, 100, timestamp, Moosique.getSequence().getResolution() / 4));
	}

	/**
	 * Draws the grid that is on the background.
	 * @param g The Graphics object used to draw the grid.
	 */
	public void paintComponent(Graphics g) {
		super.paintComponent(g);
		Graphics2D g2 = (Graphics2D)g;
		for (int c = 0; c < viewLength || c < getHeight(); c += NOTE_HEIGHT) {
			for (int r = 0; r < (10 * NOTE_WIDTH); r += NOTE_WIDTH) {
				box = new Rectangle(r, c, NOTE_WIDTH, NOTE_HEIGHT);
				g2.setColor(Color.gray);
				g2.draw(box);
			}
		}
	}
	
	/**
	 * The adapter used to listen on mouse actions
	 */
	class MAdapter extends MouseAdapter {

		/**
		 * Adds a standard note if doubleclicked.
		 */
		public void mouseClicked(MouseEvent e) {
			if (e.getClickCount() == 2) {
				popupY = e.getY();
				addStandardNote();
			}
		}
	
		public void mousePressed(MouseEvent e) {
			maybeShowPopup(e);
		}

		public void mouseReleased(MouseEvent e) {
			maybeShowPopup(e);
		}

		/**
		 * Shows the menu if an OS-specific popup-trigger was activated.
		 */
		private void maybeShowPopup(MouseEvent e) {
			if (e.isPopupTrigger()) {
				popupY = e.getY();
				popup.show(e.getComponent(), e.getX(), e.getY());
			}
		}

		/**
		 * Grabs the focus when the mouse has entered.
		 */
		public void mouseEntered(MouseEvent e) {
			// Moosique.setActiveChannel(track.getChannel());
			grabFocus();
		}
	}

	/**
	 * Listens on actions on the popupmenu and executes the appropriate action.
	 */
	class PopupListener implements ActionListener {
		public void actionPerformed(ActionEvent e) {
			Object source = e.getSource();
			if  (source == popupAdd) {
				addStandardNote();
			}
			// new MooNote(int channel, int pitch, int velocity, long timestamp, int duration)
		}
	}
}
